package MT::Plugin::AssetRenameChanger;

use strict;
use base qw( MT::Plugin );
use MT;
use MT::Util
    qw( epoch2ts encode_url format_ts relative_date perl_sha1_digest_hex);

use File::Basename;

our $VERSION = '0.01';

my $plugin = __PACKAGE__->new({
    id => 'AssetRenameChanger',
    name => 'AssetRenameChanger',
    version => $VERSION,
    description => '<MT_TRANS phrase="AssetRenameChanger.">',
    author_name => 'Yujiro Araki',
    author_link => 'http://www.koikikukan.com/',
    l10n_class  => 'AssetRenameChanger::L10N'
});
MT->add_plugin( $plugin );

*MT::CMS::Asset::js_upload_file = \&plugin_js_upload_file;

my $default_thumbnail_size = 60;

sub plugin_js_upload_file {
    my $app = shift;

    my $is_userpic = ( $app->param('type') || '' ) eq 'userpic' ? 1 : 0;
    my $user_id = $app->param('user_id');
    if ($is_userpic) {
        return $app->error(
            $app->json_error( $app->translate("Invalid Request.") ) )
            unless $user_id;

        my $user = $app->model('author')->load( { id => $user_id } )
            or return $app->error(
            $app->json_error( $app->translate("Invalid Request.") ) );

        my $appuser = $app->user;
        if ( ( !$appuser->is_superuser ) && ( $user->id != $appuser->id ) ) {
            return $app->error(
                $app->json_error( $app->translate("Permission denied.") ) );
        }
    }
    else {
        if ( my $perms = $app->permissions ) {
            return $app->error(
                $app->json_error( $app->translate("Permission denied.") ) )
                unless $perms->can_do('upload');
        }
        else {
            return $app->error(
                $app->json_error( $app->translate("Permission denied.") ) );
        }
    }
    $app->validate_magic()
        or return $app->error(
        $app->json_error( $app->translate("Invalid Request.") ) );

    # Save as asset
    my ( $asset, $bytes ) = _upload_file(
        $app,
        require_type => ( $app->param('require_type') || '' ),
        error_handler => sub {
            my ( $_app, %params ) = @_;
            return $app->error( $app->json_error( $params{error} ) );
        },
        cancel_handler => sub {
            my ( $_app, %params ) = @_;
            return $app->json_result(
                {   cancel => $app->translate(
                        "File with name '[_1]' already exists. Upload has been cancelled.",
                        $params{fname}
                    )
                }
            );
        },
    );
    return unless $asset;

    # Set tag for userpic
    if ($is_userpic) {
        $asset->tags('@userpic');
        $asset->created_by($user_id);
        $asset->save;
    }

    # Make thumbnail
    my $thumb_url;
    my $thumb_type;
    my $thumb_size = $app->param('thumbnail_size') || $default_thumbnail_size;
    if ( $asset->has_thumbnail && $asset->can_create_thumbnail ) {
        my ( $orig_height, $orig_width )
            = ( $asset->image_width, $asset->image_height );
        if ( $orig_width > $thumb_size && $orig_height > $thumb_size ) {
            ($thumb_url) = $asset->thumbnail_url(
                Height => $thumb_size,
                Width  => $thumb_size,
                Square => 1,
                Ts     => 1
            );
        }
        elsif ( $orig_width > $thumb_size ) {
            ($thumb_url)
                = $asset->thumbnail_url( Width => $thumb_size, Ts => 1 );
        }
        elsif ( $orig_height > $thumb_size ) {
            ($thumb_url)
                = $asset->thumbnail_url( Height => $thumb_size, Ts => 1 );
        }
        else {
            $thumb_url = $asset->url;
        }
        $thumb_type = 'image';
    }
    else {
        $thumb_type
            = $asset->class_type eq 'file'  ? 'default'
            : $asset->class_type eq 'video' ? 'movie'
            :                                 $asset->class_type;
    }

    # Check extension auto-change
    my $extension_message;
    if ( my $file_ext_changes = $app->param('changed_file_ext') ) {
        my ( $ext_from, $ext_to ) = split( ",", $file_ext_changes );
        $extension_message
            = $app->translate( "Extension changed from [_1] to [_2]",
            $ext_from, $ext_to )
            if ( $ext_from && $ext_to );
    }

    my $metadata = {
        id             => $asset->id,
        filename       => $asset->file_name,
        blog_id        => $asset->blog_id,
        thumbnail_type => $thumb_type,
        $thumb_url ? ( thumbnail => $thumb_url ) : (),
        ( $extension_message ? ( message => $extension_message ) : () ),
    };
    return $app->json_result( { asset => $metadata } );
}

sub _upload_file {
    my $app = shift;
    my (%upload_param) = @_;

    require MT::Image;
    my $app_id = $app->id;

    # Setup handlers
    my $eh = $upload_param{error_handler} || sub {
        start_upload(@_);
    };

    my ( $fh, $info ) = $app->upload_info('file');
    my $mimetype;
    if ($info) {
        $mimetype = $info->{'Content-Type'};
    }
    my %param = (
        entry_insert => scalar( $app->param('entry_insert') ),
        edit_field   => scalar( $app->param('edit_field') ),
        destination  => scalar( $app->param('destination') ),
        extra_path   => scalar( $app->param('extra_path') ),
        upload_mode  => $app->mode,
    );
    return $eh->(
        $app, %param,
        error => $app->translate("Please select a file to upload.")
    ) if !$fh;

    my $basename = $app->param('file') || $app->param('fname');
    $basename =~ s!\\!/!g;    ## Change backslashes to forward slashes
    $basename =~ s!^.*/!!;    ## Get rid of full directory paths
    if ( $basename =~ m!\.\.|\0|\|! ) {
        return $eh->(
            $app, %param,
            error => $app->translate( "Invalid filename '[_1]'", $basename )
        );
    }
    $basename
        = Encode::is_utf8($basename)
        ? $basename
        : Encode::decode( $app->charset,
        File::Basename::basename($basename) );

    # Change to real file extension
    if ( my $ext_new = MT::Image->get_image_type($fh) ) {
        my $ext_old
            = ( File::Basename::fileparse( $basename, qr/[A-Za-z0-9]+$/ ) )
            [2];

        if (   $ext_new ne lc($ext_old)
            && !( lc($ext_old) eq 'jpeg' && $ext_new eq 'jpg' )
            && !( lc($ext_old) eq 'swf'  && $ext_new eq 'cws' ) )
        {
            if ( $basename eq $ext_old ) {
                $basename .= '.' . $ext_new;
            }
            else {
                $basename =~ s/$ext_old$/$ext_new/;
            }
            $app->param( "changed_file_ext", "$ext_old,$ext_new" );
        }
    }

    # Setup exists/cancel handler
    my $exists_handler = $upload_param{exists_handler} || sub {
        return $eh->(
            $app, %param,
            error => $app->translate(
                "File with name '[_1]' already exists.", $basename
            )
        );
    };
    my $cancel_handler = $upload_param{cancel_handler} || sub {
        start_upload(@_);
    };

    if ( my $asset_type = $upload_param{require_type} ) {
        require MT::Asset;
        my $asset_pkg = MT::Asset->handler_for_file($basename);

        my %settings_for = (
            audio => {
                class => 'MT::Asset::Audio',
                error =>
                    $app->translate("Please select an audio file to upload."),
            },
            image => {
                class => 'MT::Asset::Image',
                error => $app->translate("Please select an image to upload."),
            },
            video => {
                class => 'MT::Asset::Video',
                error => $app->translate("Please select a video to upload."),
            },
        );

        if ( my $settings = $settings_for{$asset_type} ) {
            return $eh->( $app, %param, error => $settings->{error} )
                if !$asset_pkg->isa( $settings->{class} );
        }
    }

    my ($blog_id,        $blog,         $fmgr,
        $local_file,     $asset_file,   $base_url,
        $asset_base_url, $relative_url, $extra_path
    );
    my $label = $basename;
    if ( $blog_id = $app->param('blog_id') ) {

        $param{blog_id} = $blog_id;
        $blog = MT->model('blog')->load($blog_id)
            or return $app->error(
            $app->translate( 'Cannot load blog #[_1].', $blog_id ) );
        $fmgr = $blog->file_mgr;

        ## Build upload destination path
        my $dest = $app->param('destination');
        my $root_path;
        my $is_sitepath;
        if ( $dest =~ m/^%s/i ) {

            # sitepath
            $root_path   = $blog->site_path;
            $is_sitepath = 1;
        }
        else {
            $root_path   = $blog->archive_path;
            $is_sitepath = 0;
        }
        $dest = MT::Util::build_upload_destination($dest);

        # Make directory if not exists
        $extra_path = $app->param('extra_path') || '';
        if ($extra_path) {
            if ( $extra_path =~ m!\.\.|\0|\|! ) {
                return $eh->(
                    $app, %param,
                    error => $app->translate(
                        "Invalid extra path '[_1]'", $extra_path
                    )
                );
            }
        }
        if ( $dest ne '' ) {
            $extra_path = File::Spec->catdir( $dest, $extra_path );
        }

        my $path = File::Spec->catdir( $root_path, $extra_path );
        ## Untaint. We already checked for security holes in $relative_path.
        ($path) = $path =~ /(.+)/s;
        ## Build out the directory structure if it doesn't exist. DirUmask
        ## determines the permissions of the new directories.
        unless ( $fmgr->exists($path) ) {
            $fmgr->mkpath($path)
                or return $eh->(
                $app, %param,
                error => $app->translate(
                    "Cannot make path '[_1]': [_2]",
                    $path, $fmgr->errstr
                )
                );
        }

        # Rename filename automatically, or overwrite, or cancel upload
        {
            my $path_info = {};
            @$path_info{qw(rootPath relativePath basename)}
                = ( $root_path, $extra_path, $basename );
            my $local_file = File::Spec->catfile(
                ( $root_path, $extra_path, $basename ) );

            if ( $fmgr->exists($local_file) ) {
                my $operation_if_exists
                    = $app->param('operation_if_exists') || 0;
                if ( $operation_if_exists == 1 ) {

                    # Auto-rename
                    _rename_filename( $app, $path_info );
                }
                elsif ( $operation_if_exists == 2 ) {

                    # Overwrite, do nothing
                }
                elsif ( $operation_if_exists == 3 ) {

                    # Call cancel handler
                    return $cancel_handler->(
                        $app,
                        site_path  => $root_path,
                        extra_path => $extra_path,
                        fname      => $basename,
                    );
                }
                else {
                    # Call exists handler
                    return $exists_handler->(
                        $app,
                        site_path  => $root_path,
                        extra_path => $extra_path,
                        fname      => $basename,
                    );
                }
            }

            # Rename non-ascii filename automatically if option provided.
            if (   $app->param('auto_rename_non_ascii')
                && $path_info->{basename} =~ m/[^\x20-\x7E]/ )
            {
                # Auto-rename
                _rename_filename( $app, $path_info );
            }

            $app->run_callbacks( $app_id . '_asset_upload_path',
                $app, $fmgr, $path_info );

            ( $root_path, $extra_path, $basename )
                = @$path_info{qw(rootPath relativePath basename)};
        }

        $relative_url
            = File::Spec->catfile( $extra_path, encode_url($basename) );
        $extra_path
            = $extra_path
            ? File::Spec->catfile( $extra_path, $basename )
            : $basename;
        $asset_file = $is_sitepath ? '%r' : '%a';
        $extra_path =~ s/^[\/\\]//;
        $asset_file = File::Spec->catfile( $asset_file, $extra_path );
        $local_file = File::Spec->catfile( $path,       $basename );
        $base_url
            = $is_sitepath
            ? $blog->site_url
            : $blog->archive_url;
        $asset_base_url = $is_sitepath ? '%r' : '%a';

        ## Untaint. We have already tested $basename and $extra_path for security
        ## issues above, and we have to assume that we can trust the user's
        ## Local Archive Path setting. So we should be safe.
        ($local_file) = $local_file =~ /(.+)/s;
    }
    else {
        $blog_id        = 0;
        $asset_base_url = '%s/uploads';
        $base_url       = $app->support_directory_url . 'uploads';
        $param{support_path}
            = File::Spec->catdir( $app->support_directory_path, 'uploads' );

        require MT::FileMgr;
        $fmgr = MT::FileMgr->new('Local');
        unless ( $fmgr->exists( $param{support_path} ) ) {
            $fmgr->mkpath( $param{support_path} );
            unless ( $fmgr->exists( $param{support_path} ) ) {
                return $app->error(
                    $app->translate(
                        "Could not create upload path '[_1]': [_2]",
                        $param{support_path},
                        $fmgr->errstr
                    )
                );
            }
        }

        require File::Basename;
        my ( $stem, undef, $type )
            = File::Basename::fileparse( $basename, qr/\.[A-Za-z0-9]+$/ );

        # Rename non-ascii filename automatically if option provided.
        my $path_info = { basename => $stem };
        if (   $app->param('auto_rename_non_ascii')
            && $path_info->{basename} =~ m/[^\x20-\x7E]/ )
        {
            # Auto-rename
            _rename_filename( $app, $path_info );
        }
        my $unique_stem = $path_info->{basename};
        $local_file = File::Spec->catfile( $param{support_path},
            $unique_stem . $type );
        if ( $fmgr->exists($local_file) ) {
            my $operation_if_exists = $app->param('operation_if_exists') || 0;
            if ( $operation_if_exists == 1 ) {

                # Auto-rename
                _rename_filename( $app, $path_info );
            }
            elsif ( $operation_if_exists == 2 ) {

                # Overwrite, do nothing
            }
            elsif ( $operation_if_exists == 3 ) {

                # Call cancel handler
                return $cancel_handler->(
                    $app,
                    site_path  => $param{support_path},
                    extra_path => '',
                    fname      => $unique_stem,
                );
            }
            else {
                # Call exists handler
                return $exists_handler->(
                    $app,
                    site_path  => $param{support_path},
                    extra_path => '',
                    fname      => $unique_stem,
                );
            }
        }
        $unique_stem = $path_info->{basename};
        $local_file  = File::Spec->catfile( $param{support_path},
            $unique_stem . $type );

        my $unique_basename = $unique_stem . $type;
        $extra_path   = $unique_basename;
        $relative_url = encode_url($unique_basename);
        $asset_file
            = File::Spec->catfile( '%s', 'uploads', $unique_basename );
    }

    my ( $base, $uploaded_path, $ext )
        = File::Basename::fileparse( $basename, '\.[^\.]*' );

    if ( my $deny_exts = $app->config->DeniedAssetFileExtensions ) {
        my @deny_exts = map {
            if   ( $_ =~ m/^\./ ) {qr/$_/i}
            else                  {qr/\.$_/i}
        } split '\s?,\s?', $deny_exts;
        my @ret = File::Basename::fileparse( $basename, @deny_exts );
        if ( $ret[2] ) {
            return $app->error(
                $app->translate(
                    '\'[_1]\' is not allowed to upload by system settings.: [_2]',
                    $ext,
                    $basename
                )
            );
        }

    }

    if ( my $allow_exts = $app->config('AssetFileExtensions') ) {
        my @allow_exts = map {
            if   ( $_ =~ m/^\./ ) {qr/$_/i}
            else                  {qr/\.$_/i}
        } split '\s?,\s?', $allow_exts;
        my @ret = File::Basename::fileparse( $local_file, @allow_exts );
        unless ( $ret[2] ) {
            return $app->error(
                $app->translate(
                    '\'[_1]\' is not allowed to upload by system settings.: [_2]',
                    $ext,
                    $basename
                )
            );
        }
    }

    my ( $w, $h, $id, $write_file ) = MT::Image->check_upload(
        Fh     => $fh,
        Fmgr   => $fmgr,
        Local  => $local_file,
        Max    => $upload_param{max_size},
        MaxDim => $upload_param{max_image_dimension}
    );

    return $app->error( MT::Image->errstr )
        unless $write_file;

    ## File does not exist, or else we have confirmed that we can overwrite.
    my $umask = oct $app->config('UploadUmask');
    my $old   = umask($umask);
    defined( my $bytes = $write_file->() )
        or return $app->error(
        $app->translate(
            "Error writing upload to '[_1]': [_2]", $local_file,
            $fmgr->errstr
        )
        );
    umask($old);

    ## Close up the filehandle.
    close $fh;

    ## We are going to use $extra_path as the filename and as the url passed
    ## in to the templates. So, we want to replace all of the '\' characters
    ## with '/' characters so that it won't look like backslashed characters.
    ## Also, get rid of a slash at the front, if present.
    $extra_path =~ s!\\!/!g;
    $extra_path =~ s!^/!!;
    $relative_url =~ s!\\!/!g;
    $relative_url =~ s!^/!!;
    my $url = $base_url;
    $url .= '/' unless $url =~ m!/$!;
    $url .= $relative_url;
    my $asset_url = $asset_base_url . '/' . $relative_url;

    require File::Basename;
    my $local_basename = File::Basename::basename($local_file);
    my $local_ext
        = ( File::Basename::fileparse( $local_file, qr/\.[A-Za-z0-9]+$/ ) )
        [2];
    $local_ext =~ s/^\.//;

    require MT::Asset;
    my $asset_pkg = MT::Asset->handler_for_file($local_basename);
    my $is_image  = 0;
    if ( defined($w) && defined($h) ) {
        $is_image = 1
            if $asset_pkg->isa('MT::Asset::Image');
    }
    else {

        # rebless to file type
        $asset_pkg = 'MT::Asset'
            if $asset_pkg->isa('MT::Asset::Image');
    }
    return $app->errtrans('Uploaded file is not an image.')
        if !$is_image
        && exists( $upload_param{require_type} )
        && $upload_param{require_type} eq 'image';
    my ( $asset, $original );
    if (!(  $asset = $asset_pkg->load(
                {   class     => '*',
                    file_path => $asset_file,
                    blog_id   => $blog_id
                },
                { binary => { file_path => 1 } }
            )
        )
        )
    {
        $asset    = $asset_pkg->new();
        $original = $asset->clone;
        $asset->file_path($asset_file);
        $asset->file_name($local_basename);
        $asset->file_ext($local_ext);
        $asset->blog_id($blog_id);
        $asset->label($label);
        $asset->created_by( $app->user->id );
    }
    else {
        if ( $asset->class ne $asset_pkg->class_type ) {
            return $app->error(
                $app->translate(
                    "Cannot overwrite an existing file with a file of a different type. Original: [_1] Uploaded: [_2]",
                    $asset->class_label,
                    $asset_pkg->class_label
                )
            );
        }
        $original = $asset->clone;
        $asset->modified_by( $app->user->id );
    }
    $asset->url($asset_url);

    if ($is_image) {
        $asset->image_width($w);
        $asset->image_height($h);

        if ( $app->param('normalize_orientation') ) {
            $asset->normalize_orientation;
        }

        # Adjust image quality according to ImageQualityJpeg
        # and ImageQualityPng.
        $asset->change_quality
            if $app->config('AutoChangeImageQuality');
    }

    $asset->mime_type($mimetype) if $mimetype;
    $app->run_callbacks( $app_id . '_pre_save.asset',
        $app, $asset, $original )
        || return $app->errtrans( "Saving [_1] failed: [_2]", 'asset',
        $app->errstr );

    $asset->save;
    $app->run_callbacks( $app_id . '_post_save.asset',
        $app, $asset, $original );

    if ($is_image) {
        $app->run_callbacks(
            $app_id . '_upload_file.' . $asset->class,
            File  => $local_file,
            file  => $local_file,
            Url   => $url,
            url   => $url,
            Size  => $bytes,
            size  => $bytes,
            Asset => $asset,
            asset => $asset,
            Type  => 'image',
            type  => 'image',
            Blog  => $blog,
            blog  => $blog
        );
        $app->run_callbacks(
            $app_id . '_upload_image',
            File       => $local_file,
            file       => $local_file,
            Url        => $url,
            url        => $url,
            Size       => $bytes,
            size       => $bytes,
            Asset      => $asset,
            asset      => $asset,
            Height     => $h,
            height     => $h,
            Width      => $w,
            width      => $w,
            Type       => 'image',
            type       => 'image',
            ImageType  => $id,
            image_type => $id,
            Blog       => $blog,
            blog       => $blog
        );
    }
    else {
        $app->run_callbacks(
            $app_id . '_upload_file.' . $asset->class,
            File  => $local_file,
            file  => $local_file,
            Url   => $url,
            url   => $url,
            Size  => $bytes,
            size  => $bytes,
            Asset => $asset,
            asset => $asset,
            Type  => 'file',
            type  => 'file',
            Blog  => $blog,
            blog  => $blog
        );
    }

    return ( $asset, $bytes );
}

sub _rename_filename {
    my ( $app, $path_info ) = @_;

    my $ext = (
        File::Basename::fileparse(
            $path_info->{basename},
            qr/\.[A-Za-z0-9]+$/
        )
    )[2];

    $path_info->{basename} =~ s/(.*)$ext/$1/;
    $path_info->{basename} = $path_info->{basename} . "-" . epoch2ts( $app->blog, time ) . $ext;
}

1;
